
from enum import Enum


class GrantType(str, Enum):
    IMPLICIT = "IMPLICIT"
    PASSWORD = "PASSWORD"
    BACKEND = "BACKEND"


class PostmanProtocol(str, Enum):
    WEBSOCKET = "WEBSOCKET"
    KAFKA = "KAFKA"
    RABBITMQ = "RABBITMQ"

class ClientType(str, Enum):
    HOST = "HOST"
    CLIENT = "CLIENT"


class DataPointType(str, Enum):
    GRAPHQL = "GRAPHQL"
    REST = "REST"



class TYPENAMES(str, Enum):
    MODELPORTTYPE = "ModelPortType"



from oauthlib.oauth2.rfc6749.errors import InvalidClientError
from bergen.auths.base import AuthError, BaseAuthBackend
from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session 
from bergen.enums import ClientType
import logging
import time

logger = logging.getLogger(__name__)


class ArnheimBackendOauth(BaseAuthBackend):
    failedTries = 5
    auto_retry = True
    tokenurl_appendix = "o/token/"


    def __init__(self, host: str, port: int, client_id: str, client_secret: str, protocol="http", verify=True, **kwargs) -> None:
        super().__init__(host, port, protocol=protocol, **kwargs)
        self.client_id = client_id
        self.client_secret = client_secret
        self.protocol = protocol
        self.verify = verify
        self.base_url = f"{protocol}://{host}:{port}"       
        self.token_url = self.base_url + "/" + self.tokenurl_appendix

    def fetchToken(self, loop=None) -> str:
        # Connecting
        logger.info(f"Connecting to Arnheim at {self.token_url}")
        
        auth_client = BackendApplicationClient(client_id=self.client_id, scope=self.scope)
        oauth_session = OAuth2Session(client=auth_client, scope=self.scope)


        def fetch_token(thetry=0):
            try:

                token = oauth_session.fetch_token(token_url=self.token_url, client_id=self.client_id,
                client_secret=self.client_secret, verify=self.verify)

                if "access_token" not in token:
                    raise Exception("No access token Provided")

                return token["access_token"]
            except InvalidClientError as e:
                raise e
            except Exception as e:
                if thetry == self.failedTries or not self.auto_retry: raise AuthError(f"Cannot connect to Arnheim instance on {self.token_url}: {e}")
                logger.error(f"Couldn't connect to the Arnheim Instance at {self.token_url}. Retrying in 2 Seconds")
                time.sleep(2)
                return fetch_token(thetry=thetry + 1)
      
        return fetch_token()


    def getClientType(self) -> ClientType:
        return ClientType.EXTERNAL.value


    def getProtocol(self):
        return self.protocol


#Base Types
EXCEPTION = "exception"
ALLOWANCE = "allowance"

ASSIGNATION = "assignation"
PROVISION = "provision"
ASSIGNATION_REQUEST = "assignation_request"
PROVISION_REQUEST = "provision_request"
ACTIVATE_POD = "activate_pod"


from bergen.wards.graphql.aiohttp import AIOHttpGraphQLWard
from bergen.auths.base import BaseAuthBackend
from typing import Callable, Dict
from bergen.enums import DataPointType
from bergen.clients.base import BaseWard
import logging
import asyncio

logger = logging.getLogger(__name__)

datapointregistry = None

class DataPointRegistry(object):


    def __init__(self) -> None:
        self.pointNameWardMap: Dict[str, BaseWard] = {}
        self.builders =  {
                # Default Builders for standard
                DataPointType.GRAPHQL: lambda datapoint, bergen: AIOHttpGraphQLWard(host=datapoint.outward, port=datapoint.port, token=bergen.auth.getToken(), protocol=bergen.auth.getProtocol(), loop=bergen.loop)
        }

    def registerClientBuilder(self, type:str , builder: Callable):
        self.builders[type] = builder

    def getClientForData(self, point, auth: BaseAuthBackend) -> BaseWard:
        if point.name in self.pointNameWardMap:
            return self.pointNameWardMap[point.name]

        logger.info("Creating new DataPoint parser")

        if point.type in self.builders:
            builder = self.builders[point.type]
            self.pointNameWardMap[point.name] = builder(point, auth)
            return self.pointNameWardMap[point.name]
        else:
            raise NotImplementedError("We have no idea how to build that datatype")

    def createWardForDatapoint(self, point, bergen) -> BaseWard:
        if point.name in self.pointNameWardMap:
            return self.pointNameWardMap[point.name]

        logger.info(f"Creating new Ward for Datapoint {point}")

        if point.type in self.builders:
            builder = self.builders[point.type]
            self.pointNameWardMap[point.name]  = builder(point, bergen)
            return self.pointNameWardMap[point.name]
        else:
            raise NotImplementedError(f"We have no idea how to build the ward for this Datapoint {point.type}")

    async def configureWards(self):
        wards = [ward for point, ward in self.pointNameWardMap.items()]
        await asyncio.gather(*[ward.configure() for ward in wards])





def get_datapoint_registry() -> DataPointRegistry:
    global datapointregistry
    if datapointregistry is None:
        datapointregistry = DataPointRegistry()
    return datapointregistry


from bergen.messages.base import MessageDataModel, MessageMetaModel, MessageModel
from typing import Optional
from pydantic import BaseModel
from enum import Enum
from bergen.messages.types import EXCEPTION


class ExceptionType(str, Enum):
    CLIENT = "client"
    ARNHEIM = "arnheim"
    PROVIDER = "provider"
    POD = "pod"

class ExceptionMetaMessage(MessageMetaModel):
    type: str = EXCEPTION


class ExceptionDataMessage(MessageDataModel):
    type: ExceptionType
    base: str
    message: str
    traceback: Optional[str] 

class ExceptionMessage(MessageModel):
    data: ExceptionDataMessage
    meta: ExceptionMetaMessage





from typing import Type
from bergen.types.node.widgets.querywidget import QueryWidget
from bergen.types.node.ports.ins.base import BaseInPort
from bergen.types.model import ArnheimModel

class ModelInPort(BaseInPort):

  def __init__(self, modelClass: Type[ArnheimModel], widget=None, **kwargs) -> None:
      self.modelClass = modelClass
      if widget is None:
        meta = self.modelClass.getMeta()
        try:
            selector = meta.selector_query
            widget = QueryWidget(query=selector)
        except:
            raise Exception("You didn't provide a widget nor has it been declared in the meta class of the Model")
        
      super().__init__("model", widget, **kwargs)

  def serialize(self):
      return {**super().serialize(),"identifier" : self.modelClass.getMeta().identifier}



from bergen.types.node.widgets.base import BaseWidget


class IntWidget(BaseWidget):

    def __init__(self, min=None, max=None, **kwargs) -> None:
        super().__init__("int", **kwargs)
        self.min = min
        self.max = max
        #TODO: Inspect if widgets dependencies are okay for that query

    def params(self):
        return { "min": self.min, "max": self.max}




from bergen.query import DelayedGQL


TEMPLATE_GET_QUERY = DelayedGQL("""
query Template($id: ID,){
  template(id: $id){
    id
    node {
        id
        name
        image
        inputs {
          __typename
          key
          required
          ... on ModelPortType {
            identifier
          }
        }
        outputs {
          __typename
          key
          required
          ... on ModelPortType {
            identifier
          }
        }
    }
  }
}
""")



from bergen.types.node.widgets.intwidget import IntWidget
from bergen.types.node.ports.ins.base import BaseInPort


class IntInPort(BaseInPort):

  def __init__(self, widget = None, **kwargs) -> None:
      if widget is None:
          widget = IntWidget()
      super().__init__("int", widget,**kwargs)


from bergen.ui.widgets.base import BaseWidgetMixin
from bergen.ui.widgets.intwidget import IntWidget
from bergen.ui.widgets.querywidget import QueryWidget
from bergen.ui.widgets.sliderwidget import SliderWidget
from bergen.ui.widgets.void import VoidWidget


import json
from pydantic import BaseModel
from typing import Optional, Type, TypeVar

from pydantic.types import Json
from pydantic import Field
import uuid

class MessageMetaExtensionsModel(BaseModel):
    """ Extensions to the AMQP Message protocol

    We wrap this in its own message to allow more"""
    progress: Optional[str]

class MessageMetaModel(BaseModel):
    type: str
    reference: str = Field(default_factory=uuid.uuid4)
    extensions: Optional[MessageMetaExtensionsModel] = { }


class MessageDataModel(BaseModel):
    pass

T = TypeVar("T")

class MessageModel(BaseModel):
    data: MessageDataModel
    meta: MessageMetaModel

    def to_message(self) -> bytes:
        return json.dumps(self.dict()).encode()

    def to_channels(self) -> bytes:
        return json.dumps(self.dict())
        
    @classmethod
    def from_message(cls: Type[T], message) -> T:
        return cls(**json.loads(message.body.decode()))

    @classmethod
    def from_channels(cls: Type[T], message: str) -> T:
        return cls(**json.loads(message))

    



from abc import ABC
from bergen.wards.graphql.base import BaseGraphQLWard, GraphQLException

from gql.gql import gql
from bergen.wards.base import BaseWard
from bergen.query import GQL, TypedGQL
from gql.transport.aiohttp import AIOHTTPTransport
import logging
from gql.transport.aiohttp import log as aiohttp_logger
from gql.transport.requests import RequestsHTTPTransport


aiohttp_logger.setLevel(logging.WARNING)
import asyncio
from gql import Client, gql
import requests

class AIOHttpGraphQLWard(BaseGraphQLWard):
    can_subscribe = False

    def __init__(self, port, host, protocol, token, loop=None) -> None:
        super().__init__(port=port, host=host, protocol=protocol, token=token, loop=loop)

    async def configure(self):
        self.transport = AIOHTTPTransport(url=self._graphql_endpoint, headers=self._headers)
        await self.transport.connect()


    async def run_async(self, the_query: TypedGQL, variables: dict = {}, **kwargs):
        query_node = gql(the_query.query)

        response = await self.transport.execute(query_node, variable_values=variables)
        if response.errors:
            raise GraphQLException(str(response.errors))
        return the_query.extract(response.data)

    async def disconnect(self):
        await self.transport.close()


from bergen.query import DelayedGQL


INPUTS_FR = """
  inputs {
    __typename
      key
      required
      description
      widget {
        __typename
        dependencies
        ... on QueryWidget {
          query  
        }
      }
      label
      ... on ModelInPort {
        identifier
      }
      ... on IntInPort {
        default
      }
  }
"""


OUTPUTS_FR = """
  outputs {
    __typename
      key
      description
      ... on ModelOutPort {
        identifier
      }
  }
"""


NODE_QUERY = DelayedGQL("""
query Node($id: ID, $package: String, $interface: String){
  node(id: $id, package: $package, interface: $interface){
    id
    name
    image
""" + INPUTS_FR + """
""" + OUTPUTS_FR + """
  }
}
""")


NODE_FILTER_QUERY = DelayedGQL("""
query NodeFilter($name: String){
  nodes(name: $name){
    id
    name
    repository {
      name
    }
    description
  }
}
""")


CREATE_NODE_MUTATION = DelayedGQL("""
  mutation CreateNodeMutation($description: String!, $inputs: [InPortInputType]!, $outputs: [OutPortInputType]!, $package: String!, $interface: String!, $name: String!){
  createNode(description: $description, inputs: $inputs, outputs: $outputs, package:$package, interface: $interface, name: $name){
    id
    name
    image
""" + INPUTS_FR + """
""" + OUTPUTS_FR + """
  }
}
""")


UPDATE_OR_CREATE_NODE = DelayedGQL("""
  mutation UpdateOrCreateNode($description: String!, $inputs: [InPortInput]!, $outputs: [OutPortInput]!, $package: String!, $interface: String!, $name: String!){
  createNode(description: $description, inputs: $inputs, outputs: $outputs, package:$package, interface: $interface, name: $name){
     id
    name
    image
""" + INPUTS_FR + """
""" + OUTPUTS_FR + """
  }
}
""")




from typing import Type
from bergen.types.node.widgets.querywidget import QueryWidget
from bergen.types.node.ports.outs.base import BaseOutPort
from bergen.types.model import ArnheimModel

class ModelOutPort(BaseOutPort):

  def __init__(self, modelClass: Type[ArnheimModel],**kwargs) -> None:
      self.modelClass = modelClass
      super().__init__("model", **kwargs)

  def serialize(self):
      return {**super().serialize(),"identifier" : self.modelClass.getMeta().identifier}


from ..messages.base import MessageDataModel, MessageMetaModel, MessageMetaExtensionsModel, MessageModel

from typing import Optional
from ..messages.types import ACTIVATE_POD


class ActivatePodMetaModel(MessageMetaModel):
    type: str = ACTIVATE_POD

class ActivatePodDataModel(MessageDataModel):
    pod: Optional[int]


class ActivatePodMessage(MessageModel):
    data: ActivatePodDataModel
    meta: ActivatePodMetaModel = {"type": ACTIVATE_POD}




from oauthlib.oauth2.rfc6749.clients.legacy_application import LegacyApplicationClient
from oauthlib.oauth2.rfc6749.clients.mobile_application import MobileApplicationClient
import requests
from requests_oauthlib.oauth2_session import OAuth2Session
from bergen.auths.base import BaseAuthBackend
from bergen.enums import ClientType

class ImplicitError(Exception):
    pass


class LegacyApplication(BaseAuthBackend):


    def __init__(self, client_id = None, client_secret=None, username=None, password=None, host="localhost", port= 8000, protocol = "http", scopes= ["read"], parent=None, **kwargs) -> None:
        assert client_id is not None, "Please provide a client_id argument"
        assert client_secret is not None, "Please provide a client_secret argument"
        super().__init__(host, port, protocol=protocol, scopes=scopes)

        self.username = username
        self.password = password
        self.client_id = client_id
        self.client_secret = client_secret
        self.auth_url = self.base_url + "o/authorize"
        self.token_url = self.base_url + "o/token"

        # If you want to have a hosting QtWidget
        self.parent = parent
        self.legacy_app_client =  LegacyApplicationClient(self.client_id)
        self.scopes = scopes

        

    def fetchToken(self, loop=None) -> str:
        # Getting token
        if not self.username: self.username = input("Enter your username:    ")
        if not self.password: self.password = input("Password?               ")

        data = { "username": self.username, "password": self.password, "grant_type": "password", "scope": self.scope, "client_id": self.client_id, "client_secret": self.client_secret}

        url = self.token_url + "/"
        try:
            response = requests.post(url, data=data).json()
        except Exception as e:
            raise e


        if "access_token" in response:
            return response["access_token"]

        else:
            raise Exception(f"Wasn't authorized! {response}")


    def getClientType(self):
        return ClientType.EXTERNAL

    def getProtocol(self):
        return "http"



class BaseWidget:

  def __init__(self, type, dependencies=None) -> None:
      self.type = type
      self.dependencies = dependencies or []
      assert isinstance(self.dependencies, list), "Depencies must be a list of strings"

      super().__init__()


  def serialize(self):

      return {
          "type": self.type,
          "dependencies": self.dependencies,
          **self.params()

      }

  def params(self, port):
      return {}




from bergen.query import GQL




MARK_GQL = GQL("""
        mutation Mark($message: String!, $assignation: ID!, $level: AssignationStatus) {
            mark(message: $message, assignation: $assignation, level: $level){
                registered
            }
        }
""")


END_GQL = GQL("""
        mutation End($outputs: GenericScalar!, $assignation: ID!) {
            end(outputs: $outputs, assignation: $assignation){
                registered
            }
        }
""")


PROVISION_SUB_GQL = GQL("""
subscription Provide($node: ID!) {
  provide(node: $node, selector: "@kanal/__all__"){
    pod {
      id
    }
    status
  }
}
""")



QUEUE_GQL = GQL("""
    subscription Queue($id: ID!) {
        queue(volunteer: $id){
            id
            status
            volunteer {
                id
                node {
                    id
                    name
                }
                identifier
            }
        }
    }
""")

HOST_GQL = GQL("""
                        subscription Host($pod: ID!) {
                host(pod: $pod){
                    pod {
                        id
                    }
                    id
                    inputs
                }
                }
""")



NODE_QUERY = GQL("""
query Node($id: ID!){
  node(id: $id){
    id
    name
    image
    inputs {
      __typename
      key
      required
      ... on ModelPortType {
        identifier
      }
    }
    outputs {
      __typename
      key
      required
      ... on ModelPortType {
        identifier
      }
    }
  }
}
""")





import inspect
from typing import DefaultDict
from bergen.types.node.widgets.base import BaseWidget


class BaseOutPort:

  def __init__(self, type, key=None, label=None, description=None) -> None:
      self.type = type
      self.key = key
      self.label = label
      self.description = description
      super().__init__()


  def serialize(self):
      assert self.key is not None, "Please provide at least a key to your Port"
      return {
        "type": self.type,
        "key": self.key,
        "label" : self.label or self.key.capitalize(),
        "description": self.description,
      }


  def __call__(self, key):
      self.key = key
      return self


  def __str__(self) -> str:
      return str(self.serialize())



from bergen.types.node.widgets.intwidget import IntWidget
from bergen.types.node.ports.outs.base import BaseOutPort


class IntOutPort(BaseOutPort):

  def __init__(self, **kwargs) -> None:
      super().__init__("int",**kwargs)




from abc import ABC, abstractmethod
from bergen.messages.base import MessageModel
from re import template
from bergen.messages.activatepod import ActivatePodMessage
from typing import Union
from bergen.utils import ExpansionError, expandInputs
from bergen.messages.assignation import AssignationMessage
from bergen.schema import AssignationStatus, Template
from bergen.constants import OFFER_GQL, SERVE_GQL
from bergen.peasent.base import BaseHelper, BasePeasent
import logging
import namegenerator
import asyncio
import websockets
from bergen.models import Node
import inspect
import sys

import asyncio
try:
    from asyncio import create_task
except ImportError:
    #python 3.6 fix
    create_task = asyncio.ensure_future


logger = logging.getLogger()


class WebsocketHelper(BaseHelper):

    async def pass_yield(self, message, value):
        message.data.outputs = value
        message.data.status = AssignationStatus.YIELD
        await self.peasent.send_to_connection(message)

    async def pass_progress(self, message, value, percentage=None):
        message.data.status = AssignationStatus.PROGRESS#
        message.data.statusmessage = f'{percentage if percentage else "--"} : {value}'
        await self.peasent.send_to_connection(message)
        pass

    async def pass_result(self,message, value):
        message.data.outputs = value
        message.data.status = AssignationStatus.DONE
        await self.peasent.send_to_connection(message)

    async def pass_exception(self,message, exception):
        message.data.status = AssignationStatus.ERROR#
        message.data.statusmessage = str(exception)
        await self.peasent.send_to_connection(message)
        pass

class WebsocketPeasent(BasePeasent):
    helperClass = WebsocketHelper
    ''' Is a mixin for Our Bergen '''

    def __init__(self, *args, host= None, port= None, ssl = False, auto_reconnect=True, **kwargs) -> None:
        self.websocket_host = host
        self.websocket_port = port
        self.websocket_protocol = "wss" if ssl else "ws"


        self.auto_reconnect= auto_reconnect
        self.allowed_retries = 2
        self.current_retries = 0
        self.pod_template_map = {}

        super().__init__(*args,host=host, port=port, ssl=ssl,**kwargs)

    async def configure(self):
        self.incoming_queue = asyncio.Queue()
        self.outgoing_queue = asyncio.Queue()
        self.tasks = {}
        await self.startup()
        
    async def startup(self):
        try:
            await self.connect_websocket()
        except Exception as e:

            logger.error(f"Peasent Connection failed {e}")
            self.current_retries += 1
            if self.current_retries < self.allowed_retries and self.auto_reconnect:
                sleeping_time = (self.current_retries + 1)
                logger.info(f"Retrying in {sleeping_time} seconds")
                await asyncio.sleep(sleeping_time)
                await self.startup()
            else:
                logger.error("No reconnecting attempt envisioned. Shutting Down!")
                return

        consumer_task = create_task(
            self.consumer()
        )

        producer_task = create_task(
            self.producer()
        )

        worker_task = create_task(
            self.workers()
        )

        done, pending = await asyncio.wait(
            [consumer_task, worker_task, producer_task],
            return_when=asyncio.FIRST_EXCEPTION
        )

        logger.error(f"Lost connection inbetween everything :( {[ task.exception() for task in done]}")
        logger.error(f'Reconnecting')

        for task in pending:
            task.cancel()

        self.current_retries = 0 # reset retries after one successfull connection
        await self.startup() # Attempt to ronnect again
        

    async def connect_websocket(self):

        uri = f"{self.websocket_protocol}://{self.websocket_host}:{self.websocket_port}/provider/?token={self.token}"
        
        self.connection = await websockets.client.connect(uri)

        for pod_id, function in self.podid_function_map.items():
            activate_pod = ActivatePodMessage(data={"pod": pod_id})
            logger.info(f"Requesting to Host {activate_pod}")
            await self.send_to_connection(activate_pod)



    async def consumer(self):
        logger.warning(" [x] Awaiting Node Calls")
        async for message in self.connection:
            logger.info(f"Incoming {message}")
            await self.incoming_queue.put(message)


    async def producer(self):
        while True:
            message = await self.outgoing_queue.get()
            await self.connection.send(message.to_channels())


    async def send_to_connection(self, message: MessageModel):
        await self.outgoing_queue.put(message)
       

    async def workers(self):
        while True:
            message = await self.incoming_queue.get()
            message = AssignationMessage.from_channels(message)
            logger.info("Received Assignation")
            if message.data.status == "CANCEL":
                if message.data.reference in self.tasks: 
                    logger.info("Cancellation for task received. Canceling!")
                    task = self.tasks[message.data.reference]
                    if not task.done():
                        task.cancel()
                        logger.warn("Canceled Task!!")
                else:
                    logger.error("Received Cancellation for task that was not in our tasks..")

            else:  
                assert message.data.template is not None, "Received assignation that had no Template???"
                task = create_task(self.templateid_function_map[message.data.pod](message))
                self.tasks[message.data.reference] = task # Run in parallel


            self.incoming_queue.task_done()


    




from bergen.messages.types import PROVISION_REQUEST
from bergen.messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from pydantic import BaseModel
from typing import Optional



class ProvisionRequestParams(BaseModel):
    pass

class ProvisionRequestMetaAuthModel(MessageMetaExtensionsModel):
    token: str

class ProvisionRequestMetaExtensionsModel(MessageMetaExtensionsModel):
    progress: Optional[str]
    callback: Optional[str]

    # Set by postman
    with_progress: Optional[bool]  = False
    with_callback: Optional[bool] = False


class ProvisionRequestMetaModel(MessageMetaModel):
    type: str = PROVISION_REQUEST
    auth: ProvisionRequestMetaAuthModel
    extensions: Optional[ProvisionRequestMetaExtensionsModel]

class ProvisionRequestDataModel(MessageDataModel):
    parent: Optional[int]
    node: Optional[int] 
    template: Optional[int]
    reference: str
    params: Optional[ProvisionRequestParams]


class ProvisionRequestMessage(MessageModel):
    data: ProvisionRequestDataModel
    meta: ProvisionRequestMetaModel



from abc import ABC, abstractmethod

from requests.models import Response
from bergen.enums import ClientType
import shelve
import os
import requests
import logging


logger = logging.getLogger(__name__)



class AuthError(Exception):
    pass


class BaseAuthBackend(ABC):


    def __init__(self, host, port, protocol="http", token_url="o/token", check_endpoint="auth/", scopes=None, **kwargs) -> None:

        self.base_url = f"{protocol}://{host}:{port}/"
        self.check_url = self.base_url + check_endpoint
        self.scopes = scopes + ["introspection"] if scopes else ["introspection"]
        print(self.scopes)
        self.scope = " ".join(self.scopes)

        config_name = "access.cfg"
        run_path = os.path.abspath(os.getcwd())
        self.config_path = os.path.join(run_path, config_name)

        try:
            with shelve.open(self.config_path) as cfg:
                    self.token = cfg['token']
                    self.needs_validation = True
                    logger.debug("Found local config")
        except KeyError:
            self.token = None
            self.needs_validation = False

        super().__init__()


    @abstractmethod
    def fetchToken(self, loop=None) -> str:
        raise NotImplementedError("This is an abstract Class")



    def getToken(self, loop=None) -> str:
        if self.token is None:
            self.token = self.fetchToken()
            
            with shelve.open(self.config_path) as cfg:
                cfg['token'] = self.token

            return self.token
        
        else:
            if self.needs_validation:
                response = requests.post(self.check_url, {"token": self.token}, headers={"Authorization": f"Bearer {self.token}"})
                print(response.status_code)
                self.needs_validation = False
                if response.status_code == 200:
                    logger.info("Old token still valid!")
                    return self.token
                else:
                    logger.info("Need to refetch Token!!") # Was no longer valid, fetching anew
                    self.token = self.fetchToken()

                    with shelve.open(self.config_path) as cfg:
                        cfg['token'] = self.token

                    return self.token
                
            return self.token



    @abstractmethod
    def getClientType(self) -> ClientType:
        raise NotImplementedError("This is an Abstract Class")

    @abstractmethod
    def getProtocol(self):
        raise NotImplementedError("This is an Abstract Class")





from bergen.types.node.widgets.base import BaseWidget


class QueryWidget(BaseWidget):

    def __init__(self, query: str, **kwargs) -> None:
        super().__init__("query", **kwargs)
        self.query = query
        #TODO: Inspect if widgets dependencies are okay for that query

    def params(self):
        return { "query": self.query}



from typing import Any
from bergen.schema import AssignationParams, Pod, PodStatus, ProvisionParams
from bergen.registries.arnheim import get_current_arnheim
from bergen.types.model import ArnheimModel
from bergen.extenders.base import BaseExtender
from aiostream import stream
from bergen.extenders.contexts.pod import HostedPod
from tqdm import tqdm
import textwrap



class AssignationUIMixin:

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._ui = None


    def askInputs(self, **kwargs) -> dict:
        widget = self.getWidget(**kwargs) # We have established a ui
        if widget.exec_():
            return widget.parameters
        else:
            return None


    def getWidget(self, **kwargs):
        try:
            from bergen.ui.assignation import AssignationUI
            if not self._ui:
                self._ui = AssignationUI(self.inputs, **kwargs)
            return self._ui
        except ImportError as e:
            raise NotImplementedError("Please install PyQt5 in order to use interactive Widget based parameter query")
            
        





class NodeExtender(AssignationUIMixin, BaseExtender):


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,**kwargs)
        
        bergen = get_current_arnheim()

        self._postman = bergen.getPostman()
        self._loop, self._force_sync = bergen.getLoopAndContext()


    def provide(self, params: ProvisionParams, **kwargs):
        if self._loop.is_running() and not self._force_sync:
            return self.provide_async(params, **kwargs)
        else:
            result = self._loop.run_until_complete(self.provide_async(params, **kwargs))
            return result

    async def provide_async(self, params: ProvisionParams, **kwargs):
        return await self._postman.provide(self, params, **kwargs)       

    async def assign_async(self, inputs: dict, params: AssignationParams, **kwargs):
        
        return await self._postman.assign(self, inputs, params, **kwargs)

    async def delay_async(self, inputs: dict, params: AssignationParams, **kwargs):
    
        return await self._postman.delay(self, inputs, params, **kwargs)

    def stream(self, inputs: dict, params: AssignationParams = None, **kwargs):

        return stream.iterate(self._postman.stream(self, inputs, params, **kwargs))


    async def assign_with_progress(self, inputs, params, **kwargs):
        result = None
        with tqdm(total=100) as pbar:
                async with self.stream_progress(inputs, params, **kwargs) as stream:
                        async for item in stream:
                                result = item
                                if isinstance(result, dict): break
                                
                                progress, message = item.split(":")
                                try: 
                                        pbar.n = int(progress)
                                        pbar.refresh()
                                except:
                                        pass
                                pbar.set_postfix_str(textwrap.shorten(message, width=30, placeholder="..."))
                pbar.n = 100
                pbar.refresh()
                pbar.set_postfix_str("Done")
        return result


    def stream_progress(self,  inputs: dict, params: AssignationParams = None, **kwargs):
        return stream.iterate(self._postman.stream_progress(self, inputs, params, **kwargs)).stream()
    
    def delay(self, inputs: dict, params: AssignationParams = None, **kwargs):
        if self._loop.is_running() and not self._force_sync:
            return self.delay_async(inputs, params, **kwargs)
        else:
            result = self._loop.run_until_complete(self.delay_async(inputs, params, **kwargs))
            return result


    def __call__(self, inputs: dict, params: AssignationParams = None, with_progress = False, **kwargs) -> dict:
        """Call this node (can be run both asynchronously and syncrhounsly)

        Args:
            inputs (dict): The inputs for this Node
            params (AssignationParams, optional): [description]. Defaults to None.

        Returns:
            outputs (dict): The ooutputs of this Node
        """
    
        if self._loop.is_running() and not self._force_sync:
            if with_progress == True:
                return self.assign_with_progress(inputs, params,  with_progress=with_progress, **kwargs)
            return self.assign_async(inputs, params, with_progress=with_progress, **kwargs)

        else:
            if with_progress == True:
                return self._loop.run_until_complete(self.assign_with_progress(inputs, params,  with_progress=with_progress, **kwargs))

            result = self._loop.run_until_complete(self.assign_async(inputs, params,  with_progress=with_progress, **kwargs))
            return result



    def hosted(self, params: ProvisionParams = {}, enter_when = PodStatus.ACTIVE, **kwargs):
        return HostedPod(self, params = {}, enter_when=enter_when, **kwargs)

    def _repr_html_(self):
        string = f"{self.name}</br>"

        for input in self.inputs:
            string += "Inputs </br>"
            string += f"Port: {input._repr_html_()} </br>"

        for output in self.outputs:
            string += "Outputs </br>"
            string += f"Port: {output._repr_html_()} </br>"


        return string


from bergen.clients.base import BaseBergen
from bergen.enums import GrantType
from bergen.clients.default import Bergen
from bergen.peasent.websocket import WebsocketPeasent
import os
from bergen.auths.legacy.app import LegacyApplication
from bergen.auths.backend.app import ArnheimBackendOauth
from bergen.auths.implicit.app import ImplicitApplication


class HostBergen(WebsocketPeasent, BaseBergen):


     def __init__(self, 
    host: str = None, 
    port: int = None,
    client_id: str = None, 
    client_secret: str = None,
    grant_type: GrantType = GrantType.BACKEND,
    protocol="http", bind=True,
    allow_insecure=None,
    is_local=None,
    **kwargs) -> None:

        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1" if allow_insecure is not None else os.getenv("OAUTHLIB_INSECURE_TRANSPORT", "0")
        os.environ["ARNHEIM_LOCAL"] = "0" if is_local is not None else os.getenv("ARNHEIM_LOCAL", "0")

        protocol = protocol or os.getenv("ARNHEIM_PROTOCOL", "http")
        host = host or os.getenv("ARNHEIM_HOST", "localhost")
        port = int(port or os.getenv("ARNHEIM_PORT", "8000"))
        client_id = client_id or os.getenv("ARNHEIM_CLIENT_ID", None)
        client_secret = client_secret or os.getenv("ARNHEIM_CLIENT_SECRET", None)
        grant_type = grant_type or os.getenv("ARNHEIM_GRANT_TYPE", GrantType.BACKEND)

        herre_host = "p-tnagerl-lab1"
        herre_port = 8000

        if grant_type == GrantType.BACKEND: auth = ArnheimBackendOauth(host=herre_host, port=herre_port, client_id=client_id, client_secret=client_secret, protocol=protocol, verify=True, **kwargs)
        elif grant_type == GrantType.IMPLICIT: auth = ImplicitApplication(host=herre_host, port=herre_port, client_id=client_id, client_secret=client_secret, protocol=protocol, verify=True, **kwargs)
        elif grant_type == GrantType.PASSWORD: auth = LegacyApplication(host=herre_host, port=herre_port, client_id=client_id, client_secret=client_secret, protocol=protocol, verify=True, **kwargs)
        else: raise NotImplementedError("Please Specifiy a valid Grant Type")

        super().__init__(auth=auth, host=host, port=port, protocol = protocol, auto_negotiate=True, bind=bind, **kwargs)



from bergen.messages.types import ASSIGNATION_REQUEST
from bergen.messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from pydantic import BaseModel
from typing import Optional
from enum import Enum

class AssignationAction(str, Enum):
    CANCEL = 'cancel'
    START = "start"


class AssignationRequestParams(BaseModel):
    pass

class AssignationRequestMetaAuthModel(MessageMetaExtensionsModel):
    token: str

class AssignationRequestMetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

    # Set by postman
    with_progress: Optional[bool]  = False
    with_callback: Optional[bool] = False

class AssignationRequestMetaModel(MessageMetaModel):
    type: str = ASSIGNATION_REQUEST
    auth: AssignationRequestMetaAuthModel
    extensions: Optional[AssignationRequestMetaExtensionsModel]

class AssignationRequestDataModel(MessageDataModel):
    node: Optional[int] #TODO: Maybe not optional
    pod: Optional[int]
    template: Optional[int]
    action: AssignationAction = AssignationAction.START
    reference: str
    callback: Optional[str]
    progress: Optional[str]

    inputs: dict
    params: Optional[AssignationRequestParams]


class AssignationRequestMessage(MessageModel):
    data: AssignationRequestDataModel
    meta: AssignationRequestMetaModel